
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include "pg_crc.h"
#include "crc_sb8.h"

#ifdef _WIN32
# include <sys/timeb.h>                            /** Windows Time Handling */
#endif

#ifdef _WIN32
#define TV_GET_SECONDS(tv) (tv).time
#else
#define TV_GET_SECONDS(tv) (tv).tv_sec
#endif

#ifdef _WIN32
#define TV_GET_MILLI(tv) (tv).millitm
#else
#define TV_GET_MILLI(tv) (((tv).tv_usec)/1000)
#endif

#ifdef _WIN32
#define TV_FILL(tv) _ftime(&(tv))
#else
#define TV_FILL(tv) gettimeofday(&(tv), NULL)
#endif

#ifdef _WIN32
#define TV_DIFF_MILLI(tv1, tv2) ((tv2.time*1000+tv2.millitm)-(tv1.time*1000+tv1.millitm))
#else
#define TV_DIFF_MILLI(tv1, tv2) ((tv2.tv_sec*1000+((tv2.tv_usec)/1000))-(tv1.tv_sec*1000+((tv1.tv_usec)/1000)))
#endif

typedef struct data_block
{
  uint32    len;
  pg_crc32  crc;
  struct data_block *next;
  uint8     *data;

} DB;

static DB *head;
#define DEFAULT_BLOCK_SIZE 4 /* KBs */
#define DEFAULT_LIST_SIZE 100

/*
  command line args:

  <prog_name> <algo_name> [ <scaling_factor> [ <block_size> [ <list_size> ] ] ]

  algo_name     : PG or SB8 or BOTH (just the first char will also do!)
  scaling_factor: the multiple to use when creating the list of buffer blocks.
  block_size    : buffer size (in KBs); default is 4.
  list_size     : number of data blocks to create for the test; default is 100.

  So the command : a.out p 3 1024 300
   will test the PG's (standard) algo over 900 blocks of 1 MB each.

*/

int main( int argc, char **argv )
{
  char   algo; /* p: PG's version, s: SB8 version, b:both */
  uint32 sf; /* scaling factor */
  uint32 list_size;
  uint32 block_size; /* KBs */
  uint32 i;
  DB     *block;
  #ifdef _WIN32
  struct _timeb tv1, tv2;
  #else
  struct timeval tv1, tv2;
  #endif 

  if (argc < 2)
  {
    printf ("Please provide the algo name; PG, SB8 or BOTH\n");
    return 1;
  }

  switch (argv[1][0])
  {
  case 'P': case 'p':
    algo = 'p';
    break;
  case 'S': case 's':
    algo = 's';
    break;
  case 'B': case 'b':
    algo = 'b';
    break;
  default:
    printf ("Please provide the algo name; PG, SB8 or BOTH\n");
    return 1;
    break;
  }

  if (argc >= 3)
    sf = atoi(argv[2]);
  else
    sf = 1;

  if (argc >= 4)
    block_size = atoi(argv[3]);
  else
    block_size = DEFAULT_BLOCK_SIZE;

  if (argc >= 5)
    list_size = atoi(argv[4]);
  else
    list_size = DEFAULT_LIST_SIZE;

  printf( "test parameters:: SF=%d, block_size=%d, list_size=%d\n", sf, block_size, list_size );

  printf( "Generating data... " );
  TV_FILL( tv1 );

  for (i = 0; i < list_size * sf; ++i)
  {
    uint32 lead, body, trail;
    uint8 *p8;
    uint32 *p32;
    uint32 j;

    block = malloc(sizeof(DB) + block_size*1024);
    block->len = block_size*1024;
    block->data = ((uint8*)block) + sizeof(DB);
    block->next = head;
    head = block;

    lead = ((uint32)block->data) % sizeof(int);
    if (lead) lead = sizeof(int) - lead;

    body = ((block->len - lead)/sizeof(int))*sizeof(int);
    trail = block->len - lead - body;

    p8 = block->data;

    srand(time(NULL));
    for (j = 0; j < lead; ++j)
      *p8++ = (uint8) (255 * (rand() / (RAND_MAX + 1.0)));

    p32 = (uint32*)p8;

    for (j = 0; j < body/sizeof(int); ++j)
      *p32++ = rand();

    p8 = (uint8*)p32;

    for (j = 0; j < trail; ++j)
      *p8++ = (uint8) (255 * (rand() / (RAND_MAX + 1.0)));
  }

  TV_FILL( tv2 );
  printf( "done. Time taken : %d milliseconds\n", TV_DIFF_MILLI(tv1, tv2) );

  if (algo == 'p' || algo == 'b' )
  {
    printf( "Generating CRC using PG (vanilla) algo... " );
    TV_FILL( tv1 );

    for (block = head; block; block = block->next)
    {
      P_INIT_CRC32( block->crc );
      P_COMP_CRC32( block->crc, block->data, block->len );
      P_FIN_CRC32( block->crc );
    }

    TV_FILL( tv2 );
    printf( "done. Time taken : %d milliseconds\n", TV_DIFF_MILLI(tv1, tv2) );

    printf( "Verifying CRC using PG (vanilla) algo... " );
    TV_FILL( tv1 );

    for (block = head; block; block = block->next)
    {
      pg_crc32 crc;

      P_INIT_CRC32( crc );
      P_COMP_CRC32( crc, block->data, block->len );
      P_FIN_CRC32( crc );

      if (crc != block->crc)
        printf( "ERROR: CRC verification failed for PG version!\n" );
    }

    TV_FILL( tv2 );
    printf( "done. Time taken : %d milliseconds\n", TV_DIFF_MILLI(tv1, tv2) );

  }

  if (algo == 's' || algo == 'b' )
  {
    printf( "Generating CRC using SliceBy8 algo... " );
    TV_FILL( tv1 );

    for (block = head; block; block = block->next)
    {
      S_INIT_CRC32( block->crc );
      S_COMP_CRC32( block->crc, block->data, block->len );
      S_FIN_CRC32( block->crc );
    }

    TV_FILL( tv2 );
    printf( "done. Time taken : %d milliseconds\n", TV_DIFF_MILLI(tv1, tv2) );

    printf( "Verifying CRC using SliceBy8 algo... " );
    TV_FILL( tv1 );

    for (block = head; block; block = block->next)
    {
      pg_crc32 crc;

      S_INIT_CRC32( crc );
      S_COMP_CRC32( crc, block->data, block->len );
      S_FIN_CRC32( crc );

      if (crc != block->crc)
        printf( "ERROR: CRC verification failed for SB8 version!\n" );
    }

	TV_FILL( tv2 );
    printf( "done. Time taken : %d milliseconds\n", TV_DIFF_MILLI(tv1, tv2) );

  }

  for (block = head; block; )
  {
    DB *temp = block;
    block = block->next;
    free( temp );
  }

  return 0;
}

